/*
	xplit
	Copyright (C) 2006 Michal Vaner (michal.vaner@kdemail.net)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	he Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#define LN_SIZE 1024
#define ALLOC_ADD 1024

int debug = 0;

void deb( char *format, ... ) {
	if( debug ) {
		va_list args;
		va_start( args, format );
		vfprintf( stderr, format, args );
		va_end( args );
	}
}

enum state_t {
	cdata,
	tag,
	tag_unknown,
	attr_apo = '\'',
	attr_quot = '"',
	ignore_tag,
	tag_end,
	tag_null
} state = cdata, pre_esc;

#define WS\
	case ' ':\
	case '\n':\
	case '\r':\
	case '\t'

char *block = NULL;
int block_alloc = 0, block_size = 0;

void pushChar( char c ) {
	if( block_alloc == block_size )
		block = realloc( block, ( block_alloc += ALLOC_ADD ) * sizeof *block );
	block[ block_size ++ ] = c;
}

char *start_block, *end_block;

int level = 0;

int last_tag_start = 0;

void start_tag() {
	deb( "Starting tag, level %i\n", level + 1 );
	if( !level ) {//It is the document root
		pushChar( '>' );
		pushChar( '\0' );
		start_block = block;
		block = NULL;
		block_size = block_alloc = 0;
		char *pos = start_block + last_tag_start;//Move to the beggining of the right tag (not-ignored one)
		while( ( *pos == '<' ) || ( *pos == ' ' ) || ( *pos == '\r' ) || ( *pos == '\n' ) || ( *pos == '\t' ) )
			pos ++;
		//Now it points to the first character which is the name
		pushChar( '<' );
		pushChar( '/' );//Place the start of the ending tag
		while( ( *pos != ' ' ) && ( *pos != '\t' ) && ( *pos != '\r' ) && ( *pos != '\n' ) && ( *pos != '>' ) )
			pushChar( *( pos ++ ) );
		//Fed it with the rest
		pushChar( '>' );
		pushChar( '\0' );
		end_block = block;
		block = NULL;
		block_alloc = block_size = 0;
	}
	level ++;
}

int lc( char *c ) {
	int res = 0;
	c --;
	while( *( ++ c ) )
		res += ( *c == '\n' );
	return res;
}

void end_tag() {
	deb( "Ending tag, level %i\n", level );
	if( ( --level ) == 1 ) {
		pushChar( '>' );
		pushChar( '\0' );
		printf( "%i\n%s%s%s\n", lc( start_block ) + lc( block ) + lc( end_block ) + 1, start_block, block, end_block );
		block_alloc = 0;
		block_size = 0;
		free( block );
		block = NULL;
	}
	if( level <= 0 ) {
		fprintf( stderr, "Syntax error, too many closed tags\n" );
		exit( 1 );
	}
}

void null_tag() {
	start_tag();
	end_tag();
}

void handle_tag( char c ) {
	switch( c ) {
		case '"'://Entereing attribute value
		case '\'':
			pre_esc = state;
			state = c;
			break;
		case '>'://Ending the tag
			switch( state ) {
				case tag_unknown://Cant happen (well, <> is not valid XML)
					fprintf( stderr, "<> in XML\n" );
					exit( 1 );
				case tag://starting tag
					start_tag();
					break;
				case ignore_tag://Ignore it
					break;
				case tag_null://What about empty null-tag as a stanza?
					null_tag();
					break;
				case tag_end:
					end_tag();
					break;
				default:
					fprintf( stderr, "Can't happen, %s:%i\n", __FILE__, __LINE__ );
					exit( 2 );
			}
			state = cdata;
			break;
	}
}

void feedChar( char c ) {
	switch( state ) {
		case cdata://CData can move to tag by starting sharp bracket
			if( c == '<' ) {
				state = tag_unknown;
				last_tag_start = block_size;
			}
			break;
		case tag_unknown://An unknown tag is either ending or some other tag, determined by / after the <, or an ignore tag
			switch( c ) {
				case '/':
					state = tag_end;
					break;
				case '?':
					state = ignore_tag;
				WS:
					break;
				default:
					state = tag;
					break;
			}
			handle_tag( c );
			break;
		case tag://A tag may become temporary null-tag
			switch( c ) {
				case '/':
					state = tag_null;
					break;
			}
			handle_tag( c );
			break;
		case ignore_tag://Nothing to do with that
		case tag_end:
			handle_tag( c );
			break;
		case tag_null://Null tag is temporary, can be canceled
			switch( c ) {
				case '/':
				WS:
				case '>':
					break;
				default:
					state = tag;
			}
			handle_tag( c );
			break;
		case attr_apo://Does it end here?
		case attr_quot:
			if( c == state )
				state = pre_esc;
			break;
	}
	if( ( c != '>' ) || ( block_size ) )
		pushChar( c );
}

int main( int argc, char *argv[] ) {
	for( int i = 1; i < argc; ++i ) {
		if( !( strcmp( argv[ i ], "-h" ) && strcmp( argv[ i ], "--help" ) ) ) {
			printf( "xplit - split the XMPP stream into stanzas\n"
				"(with stream header included, so it can be fed to DOM parser)\n\n"
				"-h,--help\t\tThis help message\n"
				"-d,--debug\t\tEnable debug messages to stderr\n" );
			return 0;
		}
		if( !( strcmp( argv[ i ], "-d" ) && strcmp( argv[ i ], "--debug" ) ) )
			debug = 1;
	}
	deb( "Starting...\n" );
	int c;
	while( ( c = getchar() ) != EOF ) {
		feedChar( c );
	}
	deb( "Found an end\n" );
	return 0;
}
